"""Continuum lib.

Made by Arthur Douillard and Timothee Lesort.

The goal of this library is to provide clean and simple to use utilities for
Continual Learning.
"""
# pylint: disable=C0401
# flake8: noqa
from continuum import datasets
from continuum.scenarios import *
from continuum.tasks import *
from continuum.metrics import *
from continuum.viz import plot_samples
from continuum import generators


__version__ = "1.1.1"
